
/* Common Open Systems Environment, Common Graphics Environment	 */		
/* Extension to	PEX 5.1	version	2.  File date: February	10, 1994 */		

/******************************************************************************/
/*  (c)	Copyright Hewlett-Packard Company, 1993,  Fort Collins,	Colorado      */
/*									      */
/*  (C)	COPYRIGHT International	Business Machines Corp.	1993		      */
/*									      */
/*   Copyright (c) 1993	by Sun Microsystems, Inc.			      */
/*									      */
/*			      All Rights Reserved			      */
/*									      */
/*  Permission to use, copy, modify, and distribute this software and its     */
/*  documentation for any purpose and without fee is hereby granted,	      */
/*  provided that the above copyright notices appear in	all copies and that   */
/*  both the copyright notices and this	permission notice appear in	      */
/*  supporting documentation, and that the names of Hewlett-Packard,	      */
/*  International Business Machines, or	Sun Microsystems not be	used in	      */
/*  advertising	or publicity pertaining	to distribution	of the software	      */
/*  without specific, written prior permission.				      */
/*									      */
/*  HEWLETT-PACKARD MAKES NO WARRANTY OF ANY KIND WITH REGARD TO THIS	      */
/*  SOFTWARE, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF	      */
/*  MERCHANTABILITY AND	FITNESS	FOR A PARTICULAR PURPOSE.  Hewlett-Packard    */
/*  shall not be liable	for errors contained herein or direct, indirect,      */
/*  special, incidental	or consequential damages in connection with the	      */
/*  furnishing,	performance or use of this software.			      */
/*									      */
/*  INTERNATIONAL BUSINESS MACHINES CORPORATION	PROVIDES THIS MATERIAL	      */
/*  "AS	IS" WITHOUT WARRANTY OF	ANY KIND, EITHER EXPRESSED OR IMPLIED,	      */
/*  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY  */
/*  AND	FITNESS	FOR A PARTICULAR PURPOSE.				      */
/*									      */
/*  SUN	MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES WITH REGARD	TO THIS	      */
/*  SOFTWARE, INCLUDING	ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,*/
/*  IN NO EVENT	SHALL SUN MICROSYSTEMS,	INC. BE	LIABLE FOR ANY SPECIAL,	      */
/*  INDIRECT OR	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER	RESULTING     */
/*  FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,      */
/*  NEGLIGENCE OR OTHER	TORTIOUS ACTION, ARISING OUT OF	OR IN CONNECTION      */
/*  WITH THE USE OR PERFORMANCE	OF THIS	SOFTWARE.			      */
/*									      */
/******************************************************************************/


 /* =======================================================================
  * NOTE NOTE NOTE  Insert this	section
  *		    at the end of pl_rdr.c file
  * ====================================================================='=
  */

/***** Start of	Ext section ************/

#include "PEXExtlib.h"
#include "PEXExtlibint.h"

static void _PEXExtGenerateChangeRendererReqList ();


PEXExtRendererAttributes *
PEXExtGetRendererAttributes (display, renderer,	valueMask)

INPUT Display			*display;
INPUT PEXRenderer		renderer;
INPUT unsigned long		*valueMask;

{
    register pexExtGetRendererAttributesReq	*req;
    register PEXExtRendererAttributes		*pAttr;
    register char				*pBuf;
    pexExtGetRendererAttributesReply		*rep;
    Bool					bitSet;
    int						n, casen;
    int						count;
    int						escapeSize = 0;
    char					*escapeData;
    unsigned long				escapeOutDataSize;
    int						fpConvert;
    int						fpFormat;

    /*
     * Fill in the escape record.
     */

    escapeSize = SIZEOF	(pexExtGetRendererAttributesReq);

    escapeData = PEXAllocBuf (escapeSize);

    fpFormat = PEXGetProtocolFloatFormat (display);
    fpConvert =	(fpFormat != NATIVE_FP_FORMAT);

    req	= (pexExtGetRendererAttributesReq *) escapeData;
    req->fpFormat = fpFormat;
    req->rdr = renderer;
    req->itemMask0 = valueMask[0];
    req->itemMask1 = valueMask[1];

    /*
     * Generate	the escape.
     */

    rep	= (pexExtGetRendererAttributesReply *) PEXEscapeWithReply (display, 
	PEXExtEscapeOpcodeGetRendererAttributes, escapeSize, escapeData, 
	&escapeOutDataSize);

    PEXFreeBuf (escapeData);

    if (rep == NULL) return (NULL);	    /* return an error */

    /*
     * Start the pBuf after the	pad bytes in the reply data
     */

    pBuf = (char *) rep	+ SIZEOF (pexExtGetRendererAttributesReply);

    /*
     * Allocate	a buffer for the replies to pass back to the client.
     */

    pAttr = (PEXExtRendererAttributes *) 
	    PEXAllocBuf	(sizeof	(PEXExtRendererAttributes));

    pAttr->current_path.count =	0;
    pAttr->current_path.elements = NULL;
    pAttr->clip_list.count = 0;
    pAttr->clip_list.rectangles	= NULL;
    pAttr->pick_start_path.count = 0;
    pAttr->pick_start_path.elements = NULL;

    for	(n = 0;
	 n < (32 + (PEXExtRAMaxShift - PEXExtRAMinShift) + 1);
	 n++)
    {
	bitSet = valueMask[n >>	5] & (1L << (n & 0x1f));
	if (bitSet != 0)
	{
	    /* PEX.h uses 1L <<	n for bits in first mask word, 0x900n for
	     * bits in second word, so the switch statement cases must
	     * do the same.			   
	     */	

	    casen = (n < 32) ? 1L << (n	& 0x1f)	: PEXExtRAMinShift + n - 32;
	    switch (casen)
	    {
	    case PEXRACurrentPath:

		EXTRACT_CARD32 (pBuf, count);
		pAttr->current_path.count = count;

		pAttr->current_path.elements = (PEXElementRef *)
		    PEXAllocBuf	(count * sizeof	(PEXElementRef));

		EXTRACT_LISTOF_ELEMREF (count,
		    pBuf, pAttr->current_path.elements);
		break;

	    case PEXRAPipelineContext:

		EXTRACT_CARD32 (pBuf, pAttr->pipeline_context);
		break;

	    case PEXRAMarkerBundle:

		EXTRACT_CARD32 (pBuf, pAttr->marker_bundle);
		break;

	    case PEXRATextBundle:

		EXTRACT_CARD32 (pBuf, pAttr->text_bundle);
		break;

	    case PEXRALineBundle:

		EXTRACT_CARD32 (pBuf, pAttr->line_bundle);
		break;

	    case PEXRAInteriorBundle:

		EXTRACT_CARD32 (pBuf, pAttr->interior_bundle);
		break;

	    case PEXRAEdgeBundle:

		EXTRACT_CARD32 (pBuf, pAttr->edge_bundle);
		break;

	    case PEXRAViewTable:

		EXTRACT_CARD32 (pBuf, pAttr->view_table);
		break;

	    case PEXRAColorTable:

		EXTRACT_CARD32 (pBuf, pAttr->color_table);
		break;

	    case PEXRADepthCueTable:

		EXTRACT_CARD32 (pBuf, pAttr->depth_cue_table);
		break;

	    case PEXRALightTable:

		EXTRACT_CARD32 (pBuf, pAttr->light_table);
		break;

	    case PEXRAColorApproxTable:

		EXTRACT_CARD32 (pBuf, pAttr->color_approx_table);
		break;

	    case PEXRAPatternTable:

		EXTRACT_CARD32 (pBuf, pAttr->pattern_table);
		break;

	    case PEXRATextFontTable:

		EXTRACT_CARD32 (pBuf, pAttr->text_font_table);
		break;

	    case PEXRAHighlightIncl:

		EXTRACT_CARD32 (pBuf, pAttr->highlight_incl);
		break;

	    case PEXRAHighlightExcl:

		EXTRACT_CARD32 (pBuf, pAttr->highlight_excl);
		break;

	    case PEXRAInvisibilityIncl:

		EXTRACT_CARD32 (pBuf, pAttr->invisibility_incl);
		break;

	    case PEXRAInvisibilityExcl:

		EXTRACT_CARD32 (pBuf, pAttr->invisibility_excl);
		break;

	    case PEXRARendererState:

		EXTRACT_LOV_CARD16 (pBuf, pAttr->renderer_state);
		break;

	    case PEXRAHLHSRMode:

		EXTRACT_LOV_INT16 (pBuf, pAttr->hlhsr_mode);
		break;

	    case PEXRANPCSubVolume:

		EXTRACT_NPC_SUBVOLUME (pBuf, pAttr->npc_subvolume,
		    fpConvert, fpFormat);
		break;
		
	    case PEXRAViewport:

		EXTRACT_VIEWPORT (pBuf,	pAttr->viewport, fpConvert, fpFormat);
		break;

	    case PEXRAClipList:

		EXTRACT_CARD32 (pBuf, count);
		pAttr->clip_list.count = count;

		pAttr->clip_list.rectangles = (PEXDeviceRect *)
		    PEXAllocBuf	(count * sizeof	(PEXDeviceRect));

		EXTRACT_LISTOF_DEVRECT (count,
		    pBuf, pAttr->clip_list.rectangles);
		break;

	    case PEXRAPickIncl:

		EXTRACT_CARD32 (pBuf, pAttr->pick_incl);
		break;

	    case PEXRAPickExcl:

		EXTRACT_CARD32 (pBuf, pAttr->pick_excl);
		break;

	    case PEXRAPickStartPath:

		EXTRACT_CARD32 (pBuf, count);
		pAttr->pick_start_path.count = count;

		pAttr->pick_start_path.elements	= (PEXElementRef *)
		    PEXAllocBuf	(count * sizeof	(PEXElementRef));

		EXTRACT_LISTOF_ELEMREF (count,
		    pBuf, pAttr->pick_start_path.elements);
		break;

	    case PEXRABackgroundColor:

		EXTRACT_COLOR_SPEC (pBuf, pAttr->background_color,
		    fpConvert, fpFormat);
		break;

	    case PEXRAClearImage:

		EXTRACT_LOV_CARD8 (pBuf, pAttr->clear_image);
		break;

	    case PEXRAClearZ:

		EXTRACT_LOV_CARD8 (pBuf, pAttr->clear_z);
		break;

	    case PEXRAEchoMode:

		EXTRACT_LOV_CARD16 (pBuf, pAttr->echo_mode);
		break;

	    case PEXExtRATMBindingTable:

		EXTRACT_CARD32 (pBuf, pAttr->tm_binding_table);
		break;

	    case PEXExtRATMCoordSourceTable:	       

		EXTRACT_CARD32 (pBuf, pAttr->tm_coord_source_table);
		break;

	    case PEXExtRATMCompositionTable:

		EXTRACT_CARD32 (pBuf, pAttr->tm_composition_table);
		break;

	    case PEXExtRATMSamplingTable:

		EXTRACT_CARD32 (pBuf, pAttr->tm_sampling_table);
		break;

	    default:	       /* handle unused	bits in	mask */
		break;
	    }
	}
    }

    /*
     * Done, so	free the Escape	call reply data.
     */

    Xfree (rep);
    return (pAttr);
}


void
PEXExtChangeRenderer (display, renderer, valueMask, values)

INPUT Display			*display;
INPUT PEXRenderer		renderer;
INPUT unsigned long		*valueMask;
INPUT PEXExtRendererAttributes	*values;

{
    register pexExtChangeRendererReq	*req;
    int					escapeSize;
    char				*escapeData;
    int					fpConvert;
    int					fpFormat;

    /*
     * For the item mask,
     * turn off	the Current Path and Renderer State attributes,	since
     * these are not modifiable.
     */

    valueMask[0] &= ~(PEXRACurrentPath | PEXRARendererState);

    /*
     * Fill in the escape record.
     */

    fpFormat = PEXGetProtocolFloatFormat (display);
    fpConvert =	(fpFormat != NATIVE_FP_FORMAT);

    _PEXExtGenerateChangeRendererReqList (display, fpConvert, fpFormat,
	valueMask, values, &escapeSize,	&escapeData);

    req	= (pexExtChangeRendererReq *) escapeData;
    req->fpFormat = fpFormat;
    req->rdr = renderer;
    req->itemMask0 = valueMask[0];
    req->itemMask1 = valueMask[1];

    /*
     * Generate	the escape.
     */

    PEXEscape (display,	PEXExtEscapeOpcodeChangeRenderer,
	       escapeSize, escapeData);

    PEXFreeBuf (escapeData);
}


/*
 * Routine to write a packed list of renderer attributes into
 * a ExtChangeRendererReq data structure in a buffer.
 */

static void
_PEXExtGenerateChangeRendererReqList (display, fpConvert, fpFormat,
    valueMask, values, sizeRet,	listRet)

INPUT Display			*display;
INPUT int			fpConvert;
INPUT int			fpFormat;
INPUT unsigned long		*valueMask;
INPUT PEXExtRendererAttributes	*values;
OUTPUT int			*sizeRet;
OUTPUT char			**listRet;

{
    register char	*pBuf;
    int			size;
    Bool		bitSet;
    int			n, casen;

    /*
     * size is the maximum size	we might need to store the Renderer list.  Just
     * use 2*sizeof(PEXExtRendererAttributes) to account for padding between 
     * shorts.
     * And allow room for the PEXExtChangeRenderer Escape header.	   
     */

    size = 2 * sizeof (PEXExtRendererAttributes) +
	SIZEOF (pexExtChangeRendererReq);

    if (valueMask[0] & PEXRAClipList)
    {
	size +=	(values->clip_list.count * SIZEOF (pexDeviceRect));
    }

    if (valueMask[0] & PEXRAPickStartPath)
    {
	size +=	(values->pick_start_path.count * SIZEOF	(pexElementRef));
    }

    *listRet = (char *)	PEXAllocBuf (size);
    pBuf = *listRet + SIZEOF (pexExtChangeRendererReq);

    /*
     * Pack the	attributes.
     */

    for	(n = 0;
	 n < (32 + (PEXExtRAMaxShift - PEXExtRAMinShift) + 1);
	 n++)
    {
	bitSet = valueMask[n >>	5] & (1L << (n & 0x1f));
	if (bitSet != 0)
	{
	    /* PEX.h uses 1L <<	n for bits in first mask word, 0x900n for
	     * bits in second word, so the switch statement cases must
	     * do the same.			   
	     */	   

	    casen = (n < 32) ? 1L << (n	& 0x1f)	: PEXExtRAMinShift + n - 32;
	    switch (casen)
	    {
	    case PEXRACurrentPath:

		/*
		 * Current path	doesn't	make sense in a	new or changed
		 * renderer, so	ignore it.
		 */
		break;

	    case PEXRAPipelineContext:

		STORE_CARD32 (values->pipeline_context,	pBuf);
		break;

	    case PEXRAMarkerBundle:

		STORE_CARD32 (values->marker_bundle, pBuf);
		break;

	    case PEXRATextBundle:

		STORE_CARD32 (values->text_bundle, pBuf);
		break;

	    case PEXRALineBundle:

		STORE_CARD32 (values->line_bundle, pBuf);
		break;

	    case PEXRAInteriorBundle:

		STORE_CARD32 (values->interior_bundle, pBuf);
		break;

	    case PEXRAEdgeBundle:

		STORE_CARD32 (values->edge_bundle, pBuf);
		break;

	    case PEXRAViewTable:

		STORE_CARD32 (values->view_table, pBuf);
		break;

	    case PEXRAColorTable:

		STORE_CARD32 (values->color_table, pBuf);
		break;

	    case PEXRADepthCueTable:

		STORE_CARD32 (values->depth_cue_table, pBuf);
		break;

	    case PEXRALightTable:

		STORE_CARD32 (values->light_table, pBuf);
		break;

	    case PEXRAColorApproxTable:

		STORE_CARD32 (values->color_approx_table, pBuf);
		break;

	    case PEXRAPatternTable:

		STORE_CARD32 (values->pattern_table, pBuf);
		break;

	    case PEXRATextFontTable:

		STORE_CARD32 (values->text_font_table, pBuf);
		break;

	    case PEXRAHighlightIncl:

		STORE_CARD32 (values->highlight_incl, pBuf);
		break;

	    case PEXRAHighlightExcl:

		STORE_CARD32 (values->highlight_excl, pBuf);
		break;

	    case PEXRAInvisibilityIncl:

		STORE_CARD32 (values->invisibility_incl, pBuf);
		break;

	    case PEXRAInvisibilityExcl:

		STORE_CARD32 (values->invisibility_excl, pBuf);
		break;

	    case PEXRARendererState:

		/*
		 * Renderer state doesn't make sense in	a new or changed
		 * renderer, so	ignore it.
		 */
		break;

	    case PEXRAHLHSRMode:

		STORE_CARD32 (values->hlhsr_mode, pBuf);
		break;

	    case PEXRANPCSubVolume:

		STORE_NPC_SUBVOLUME (values->npc_subvolume, pBuf,
		    fpConvert, fpFormat);
		break;

	    case PEXRAViewport:

		STORE_VIEWPORT (values->viewport, pBuf,	fpConvert, fpFormat);
		break;

	    case PEXRAClipList:

		STORE_CARD32 (values->clip_list.count, pBuf);

		STORE_LISTOF_DEVRECT (values->clip_list.count,
		    values->clip_list.rectangles, pBuf);
		break;

	    case PEXRAPickIncl:

		STORE_CARD32 (values->pick_incl, pBuf);
		break;

	    case PEXRAPickExcl:

		STORE_CARD32 (values->pick_excl, pBuf);
		break;

	    case PEXRAPickStartPath:

		STORE_CARD32 (values->pick_start_path.count, pBuf);

		STORE_LISTOF_ELEMREF (values->pick_start_path.count,
		    values->pick_start_path.elements, pBuf);
		break;

	    case PEXRABackgroundColor:

		STORE_COLOR_SPEC (values->background_color, pBuf,
		    fpConvert, fpFormat);
		break;

	    case PEXRAClearImage:

		STORE_CARD32 (values->clear_image, pBuf);
		break;

	    case PEXRAClearZ:

		STORE_CARD32 (values->clear_z, pBuf);
		break;

	    case PEXRAEchoMode:

		STORE_CARD32 (values->echo_mode, pBuf);
		break;

	    case PEXExtRATMBindingTable:

		STORE_CARD32 (values->tm_binding_table,	pBuf);
		break;

	    case PEXExtRATMCoordSourceTable:	 

		STORE_CARD32 (values->tm_coord_source_table, pBuf);
		break;

	    case PEXExtRATMCompositionTable:

		STORE_CARD32 (values->tm_composition_table, pBuf);
		break;

	    case PEXExtRATMSamplingTable:

		STORE_CARD32 (values->tm_sampling_table, pBuf);
		break;

	    default:	       /* handle unused	bits in	mask */
		break;
	    }
	}
    }

    *sizeRet = pBuf - *listRet;
}
/***** End of Ext section ************/
